;; Generated package description from eldoc-mouse.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "eldoc-mouse" "3.0.7" "Display documentation for mouse hover" '((emacs "27.1") (posframe "1.4.0") (eglot "1.8")) :commit "67bb73e54533e276e35d2548aa30e299c3a5ee18" :authors '(("Huang Feiyu" . "sibadake1@163.com")) :maintainer '("Huang Feiyu" . "sibadake1@163.com") :keywords '("tools" "languages" "convenience" "mouse" "hover") :url "https://github.com/huangfeiyu/eldoc-mouse")
